Chat = Core.class(Sprite)

Chat.default = {	
	width = 640,
	height = 320,
	
	nick = "???",
	
	logFile = "|D|chat.log",
	maxLogSize = 100000,
	infoSize = 20,

	okText = "Ok",
	cancelText = "Cancel",

	fontName = app.fontName,
	boldEffect = 0.06,
	fontHeight = 40,

	cellHeight = 40,
	minCellHeight = 20,
	maxCellHeight = 100,
	
	indent = 0.1,
	messageInterval = 3,

	scrollDeltaDumping = 0.6,
	scrollDragDumping = 0.6,
	scrollFreeDumping = 0.95,
	
	holdDistance = 10,
	holdTime = 0.2,
	
	frameColor = 0xFFFFFF,
	frameAlpha = 0.0,
	
	ownColors = {
		nick = 0xFF3333,
		time = 0x888888,
		text = 0x000000,
		bg   = 0x77FF22,
	},
	
	foreignColors = {
		nick = 0x333311,
		time = 0x888888,
		text = 0x000000,
		bg   = 0x66FFFF,
	},
	
	blockAlpha = 0.8,
	bgColor = 0x00FFFF,
	bgAlpha = 0.1,
}

function Chat:init(p)
	for k,v in pairs(self.default) do self[k] = p[k] or v end

	self.font = app.loadFont(app.fontName, self.fontHeight, self.cellHeight)
	self.log = self:loadLog()

	self.pointerY0, self.pointerY1, self.pointerY = 0, 0, 0
	self.pointerT0 = 0
	self.pointerDist = 0
	self.pointerDrag = false
	self.scrollAcc = 0

	self.stamp = TextField.new(self.font)
	self.bgStamp = Rectangle.new()

	self.rectangle = Rectangle.new(self.width, self.height,
		self.bgColor, self.bgAlpha)
	self:addChild(self.rectangle)

	self.render = RenderTarget.new(self.width, self.height, true)
	self.sprite = Bitmap.new(self.render)
	self:addChild(self.sprite)

	self.frame = Frame.new(self.width-1, self.height-1, self.frameColor, self.frameAlpha)
	self:addChild(self.frame)

	self:calcRenderSettings()
	self:renderText()
	
	self:addEventListener(Event.MOUSE_DOWN, self.onMouseDown, self)
	self:addEventListener(Event.MOUSE_MOVE, self.onMouseMove, self)
	self:addEventListener(Event.MOUSE_UP, self.onMouseUp, self)

	self:addEventListener(Event.TOUCHES_BEGIN, self.onTouchesBegin, self)
	self:addEventListener(Event.TOUCHES_MOVE, self.onTouchesMove, self)
	self:addEventListener(Event.TOUCHES_END, self.onTouchesEnd, self)
	self:addEventListener(Event.TOUCHES_CANCEL, self.onTouchesCancel, self)
	
	self:addEventListener(Event.ENTER_FRAME, self.onEnterFrame, self)
	
	self:addEventListener(Event.KEY_DOWN, self.inputMessage, self)
	
	self:addEventListener(Event.APPLICATION_EXIT, self.saveLog, self)
	self:addEventListener(Event.REMOVED_FROM_STAGE, self.saveLog, self)
	self:addEventListener(Event.APPLICATION_SUSPEND, self.saveLog, self)
end

function Chat:onMouseDown(event)
---[[
	if self.rectangle:hitTestPoint(event.x, event.y) then
		self.focus = true
		self.pointerDrag = false
		self.pointerRelease = false
		self.pointerY0, self.pointerY1, self.pointerY =
			event.y, event.y, event.y
		self.pointerT0 = os.timer()
		self.pointerDist = 0
	end
--]]
end

function Chat:onMouseMove(event)
	if self.focus then self.pointerX, self.pointerY = event.x, event.y end
end

function Chat:onMouseUp(event)
	if self.focus and not self.pointerDrag then self:inputMessage() end
	self.focus = false
	
end

function Chat:onTouchesBegin(event)
	event.x, event.y = event.touch.x, event.touch.y
	self:onMouseDown(event)
end

function Chat:onTouchesMove(event)
	event.x, event.y = event.touch.x, event.touch.y
	self:onMouseMove(event)
end

function Chat:onTouchesEnd(event)
	self:onMouseUp(event)
end

function Chat:onTouchesCancel(event)
	self:onMouseUp(event)
end

function Chat:onEnterFrame()
	local dY = self.pointerY - self.pointerY1
	
	if self.focus then
		self.scrollAcc = self.scrollDragDumping * (
			self.scrollAcc + self.scrollDeltaDumping * dY)
		self.pointerDist = self.pointerDist + dY
		
		if not self.pointerDrag and
			(math.abs(self.pointerDist) > self.holdDistance or
				os.timer() - self.pointerT0 > self.holdTime)
		then self.pointerDrag = true end
	else
		if self.pointerDrag then
			self.scrollAcc = self.scrollFreeDumping * self.scrollAcc
			if math.abs(self.scrollAcc) < 1 then self.scrollAcc = 0 end
			dY = dY + self.scrollAcc
			if dY == 0 then self.pointerDrag = false end
		end
	end
	
	if not self.pointerDrag then dY = 0 end

	if self.slidingHeight < 0 and dY ~= 0 then
		local y = self.offset + dY
		if y > 0 then
			y = 0
			self.scrollAcc = 0
		elseif y < self.slidingHeight then
			y = self.slidingHeight
			self.scrollAcc = 0
		end
		self.offset = y
		self:renderText()
	end
	
	self.pointerY1 = self.pointerY
end

function Chat:onResize(width, height)
	self.render = RenderTarget.new(width, height, true)
	collectgarbage()
	self.sprite:setTexture(self.render)
	
	self.rectangle:setSize(width, height)
	self.frame:setSize(width-1, height-1)
	
	self:calcRenderSettings()
	self:renderText()

	self.focus = false
end

function Chat:loadLog()
	local file = io.open(self.logFile)
	if not file then file = io.open(self.logFile, "wb") end
	local data = file:read("*a")
	file:close()
	local isLog, logData = pcall(json.decode, data)
	self.logSize = isLog and #data or 0

	if not isLog then return {} end

	local log = {}
	for k,v in ipairs(logData) do
		table.insert(log, {
			nick = v[1],
			time = v[2],
			text = v[3],
		})
	end
	
	return log
end

function Chat:saveLog()
	local file = io.open(self.logFile, "wb")
	local t = {}
	for k,v in ipairs(self.log) do
		table.insert(t, {v.nick, v.time, v.text})
	end
	local data = json.encode(t)
	file:write(data)
	file:close()
end

function Chat:addMessage(nick, text)
	table.insert(self.log, {nick = nick, time = os.time(), text = text})
	local l = #self.log
	local msg = self.log[l]
	msg.rows = fitText(text, self.fitWidth, self.font)
	msg.bgnY = l > 1 and self.log[l-1].endY + self.messageInterval or 0
	msg.endY = msg.bgnY + self.cellHeight * (#msg.rows+1)
	self.logSize = self.logSize + #nick + #text + self.infoSize
	if self.logSize > self.maxLogSize then
		while self.logSize > self.maxLogSize do
			local msg = self.log[1]
			table.remove(self.log, 1)
			self.logSize = self.logSize - #msg.nick - #msg.text - self.infoSize
		end
		self:calcRenderSettings()
	end

	self:calcSlidingHeight()
	self.offset = self.slidingHeight
	self:renderText()
end

function Chat:calcSlidingHeight()
	self.slidingHeight = math.min(0, self.height -
		(self.log[#self.log] and self.log[#self.log].endY or 0))
end

function Chat:calcRenderSettings()
	self.pixelIndent = math.floor(self.indent * self.cellHeight)
	self.fitWidth = self.width - 2 * self.pixelIndent
	
	self.log[0] = {endY = - self.messageInterval}
	for k,v in ipairs(self.log) do
		v.rows = fitText(v.text, self.fitWidth, self.font)
		v.bgnY = self.log[k-1].endY + self.messageInterval
		v.endY = v.bgnY + (1 + #v.rows) * self.cellHeight
	end
	self.log[0] = nil
	
	self:calcSlidingHeight()
	
	self.offset = self.offset or 0
	if self.offset < self.slidingHeight then self.offset = self.slidingHeight end
end


function Chat:renderText()
	local sliderEvent = Event.new "scroll"
	sliderEvent.value = self.offset / self.slidingHeight
	sliderEvent.size = self.height / (self.height - self.slidingHeight)
	self:dispatchEvent(sliderEvent)

	self.render:clear(self.bgColor, self.bgAlpha)
	
	if #self.log < 1 then return end
	
	local pos1 = nil
	do
		local y = - self.offset
		for k,v in ipairs(self.log) do
			if v.bgnY > y then pos1 = k - 1; break end
		end
		pos1 = (pos1 and pos1 > 1) and pos1 or 1
	end

	for i = pos1, #self.log do
		local m = self.log[i]
		local isMe = m.nick == ""
		local colors = isMe and self.ownColors or self.foreignColors

		self.bgStamp:setX(0)
		self.bgStamp:setY(m.bgnY + self.offset)
		self.bgStamp:setSize(self.width, self.cellHeight * (1 + #m.rows))
		self.bgStamp:setColor(colors.bg, self.blockAlpha)
		self.render:draw(self.bgStamp)
		
		self.stamp:setY(m.bgnY + self.font.offset + self.cellHeight + self.offset)
			
		self.stamp:setX(self.pixelIndent)
		self.stamp:setTextColor(colors.nick)
		self.stamp:setText(isMe and self.nick or m.nick)
		self.render:draw(self.stamp)
		
		if self.boldEffect > 0 then
			self.stamp:setX(self.pixelIndent +
				math.floor(self.fontHeight*self.boldEffect))
			self.render:draw(self.stamp)
		end
		
		local nickWidth = self.stamp:getWidth()
		
		self.stamp:setTextColor(colors.time)
		self.stamp:setText(os.date("%c", m.time))
		local timeWidth = self.stamp:getWidth()
		if nickWidth + timeWidth + self.cellHeight < self.width then
			self.stamp:setX(self.width - timeWidth - self.pixelIndent)
			self.render:draw(self.stamp)
		else
			self.stamp:setText(os.date("%X", m.time))
			local timeWidth = self.stamp:getWidth()
			if nickWidth + timeWidth + self.cellHeight < self.width then
				self.stamp:setX(self.width - timeWidth - self.pixelIndent)
				self.render:draw(self.stamp)
			end
		end

		for k,v in ipairs(m.rows) do
			self.stamp:setTextColor(colors.text)
			self.stamp:setX(self.pixelIndent)
			local y = m.bgnY + (k + 1) * self.cellHeight +
				self.offset + self.font.offset
			if y > self.height + self.cellHeight then return end
			self.stamp:setY(y)
			self.stamp:setText(v)
			self.render:draw(self.stamp)
		end
	end
end

function Chat:scrollTo(v)
	self.offset = v * self.slidingHeight
	self:renderText()
end

function Chat:inputMessage()
	local dialog = TextInputDialog.new(app.T "New Message", "", "",
		app.T(self.cancelText), app.T(self.okText))
	dialog:addEventListener(Event.COMPLETE, function(event)
		if event.buttonIndex ~= 1 or event.text == "" then return end
		local messageEvent = Event.new "message"
		messageEvent.text = event.text
		self:dispatchEvent(messageEvent)
		--self:addMessage("", event.text)
	end)
	dialog:show()
end

function Chat:setCellHeight(h)
	if self.minCellHeight <= h and h <= self.maxCellHeight then
		self.cellHeight = h
		self:onResize(self.width, self.height)
	end
end